#include <stdlib.h>
#include <stdio.h>
#include <assert.h>
#include <math.h>

#include "outstreams.hpp"
#include "coordinates.hpp"
#include "SC2Map.hpp"


float SC2Map::influenceWeightGround = 0.55f;
float SC2Map::influenceWeightCWalk  = 0.10f;
float SC2Map::influenceWeightAir    = 0.35f;

void SC2Map::setInfluenceWeightGround( float iw )
{
  influenceWeightGround = iw;
}

void SC2Map::setInfluenceWeightCWalk( float iw )
{
  influenceWeightCWalk = iw;
}

void SC2Map::setInfluenceWeightAir( float iw )
{
  influenceWeightAir = iw;
}


void SC2Map::identifyBases()
{
  nameStartLocs();
  locateBases();
}


void SC2Map::analyzeBases()
{
  calculateInfluence();
  assignMainNatThirdIslands();
  calculatePositionalBalance();
}


void SC2Map::locateBases()
{
  for( list<Resource*>::const_iterator itr = resources.begin();
       itr != resources.end();
       ++itr )
  {
    Resource* r = *itr;

    Base* b;

    if( !getNearestBase( r, &b ) )
    {
      // skip resources placed on unpathable areas
      continue;
    }

    if( b == NULL )
    {
      // start a new base for this resource
      b = new Base();
      b->loc.set( &(r->loc) );

      b->resources.push_back( r );
      b->resourceTotal = r->amount;

      this->bases.push_back( b );

    } else {
      // there is already a base for this
      // resource, add it and calculate new
      // position for the base
      b->resources.push_back( r );
      b->resourceTotal += r->amount;

      float mxTotal = 0;
      float myTotal = 0;
      for( list<Resource*>::const_iterator itr = b->resources.begin();
           itr != b->resources.end();
           ++itr )
      {
        mxTotal += (*itr)->loc.mx;
        myTotal += (*itr)->loc.my;
      }
      mxTotal /= b->resources.size();
      myTotal /= b->resources.size();

      // check point to see if its over ground...
      point p;
      p.mSet( mxTotal, myTotal );
      Node* n = getPathNode( &p, PATH_GROUND_NOROCKS );

      // only update the base's location if this is true
      if( n != NULL )
      {
        b->loc.pcSet( p.pcx, p.pcy );
      }
    }
  }
}


// if the resource is over unpathable cells, say as a decoration
// instead of a real resource, and there is no path node for it,
// return FALSE and ignore it in the identification of bases
bool SC2Map::getNearestBase( Resource* r, Base** bOut )
{
  Node* nResource = getPathNode( &(r->loc), PATH_GROUND_WITHROCKS );

  if( nResource == NULL )
  {
    return false;
  }

  Base* b = NULL;
  float d = infinity;

  for( list<Base*>::iterator itr = bases.begin();
       itr != bases.end();
       ++itr )
  {
    Base* bConsider = *itr;
    Node* nBase     = getPathNode( &(bConsider->loc), PATH_GROUND_WITHROCKS );
    float dConsider = getShortestPathDistance( nResource, nBase, PATH_GROUND_WITHROCKS );

    // 25 units is as far as a resource in a base should be from
    // any other resource that belongs to the same base
    if( dConsider < d && dConsider < 25.0f )
    {
      b = bConsider;
      d = dConsider;
    }
  }

  *bOut = b;
  return true;
}


void SC2Map::nameStartLocs()
{
  // atan2 returns an angle in radians in range [-pi, pi]
  // so let's cut the range into 13 parts, where we will find
  // half of 9 o'clock at the bottom, and half of it at the top.
  float upperBounds[13];

  // each wedge of clock is 1/12 of 2*pi, or pi/6
  float wedge = M_PI / 6.0f;

  // the upper bound of 9 o'clock's BOTTOM half is half a wedge
  // over the bottom of the range
  float bound = -M_PI + 0.5f*wedge;

  for( int i = 0; i < 13; ++i )
  {
    upperBounds[i] = bound;
    bound += wedge;
  }

  // start at bottom of range, go positive radians around clock
  char wedgeNames[][16] =
  {
    "9 o'clock",
    "8 o'clock",
    "7 o'clock",
    "6 o'clock",
    "5 o'clock",
    "4 o'clock",
    "3 o'clock",
    "2 o'clock",
    "1 o'clock",
    "12 o'clock",
    "11 o'clock",
    "10 o'clock",
    "9 o'clock",
  };

  char wedgeNumbers[13] = { 9, 8, 7, 6, 5, 4, 3, 2, 1, 12, 11, 10, 9 };

  // Try to use this scheme unless two start locations have
  // the same number--then give up and just number them
  bool clockNumUsed[13] = { false };
  bool reusedNum        = false;

  // set a point at the playable terrain's center
  // to calculate the angle to a start location
  point center;
  center.ptSet( txDimPlayable/2,
                tyDimPlayable/2 );

  for( list<StartLoc*>::const_iterator itr = startLocs.begin();
       itr != startLocs.end();
       ++itr )
  {
    StartLoc* sl = *itr;

    // TODO alter angle for aspect ratio?
    float ang = atan2( sl->loc.my - center.my,
                       sl->loc.mx - center.mx );

    for( int i = 0; i < 13; ++i )
    {
      if( ang < upperBounds[i] )
      {
        // find out if the clock wedge is already used,
        if( clockNumUsed[i] ) { reusedNum = true; }

        // if we are at 9, mark BOTH halves as used
        if( i == 0 || i == 12 )
        {
          clockNumUsed[0]  = true;
          clockNumUsed[12] = true;
        } else {
          clockNumUsed[i]  = true;
        }

        strcpy( sl->name, wedgeNames[i] );
        sl->idNum = wedgeNumbers[i];

        // done looking for a wedge for this start loc
        break;
      }
    }
  }

  if( reusedNum )
  {
    // that scheme didn't work out, just number them
    int n = 1;

    for( list<StartLoc*>::const_iterator itr = startLocs.begin();
         itr != startLocs.end();
         ++itr )
    {
      StartLoc* sl = *itr;
      sprintf( sl->name, "%d", n );
      sl->idNum = n;
      ++n;
    }
  }
}


void SC2Map::calculateInfluence()
{
  // THIS CALCULATION HAS CHANGED--influence is calculated
  // separately for EACH PAIR of start locations, so we
  // are considering all the spawn configurations

  // initialize this calc
  for( list<StartLoc*>::const_iterator slItr1 = startLocs.begin();
       slItr1 != startLocs.end();
       ++slItr1 )
  {
    StartLoc* sl1 = *slItr1;

    for( list<StartLoc*>::const_iterator slItr2 = startLocs.begin();
         slItr2 != startLocs.end();
         ++slItr2 )
    {
      StartLoc* sl2 = *slItr2;

      if( sl1 == sl2 ) { continue; }

      sl1->sl2resourceInfluence[sl2] = 0.0f;
    }
  }

  // how much influence, as a percentage, does each
  // start location IN A PAIR exert on a given base?
  for( list<Base*>::const_iterator bItr = bases.begin();
       bItr != bases.end();
       ++bItr )
  {
    Base* b = *bItr;

    for( list<StartLoc*>::const_iterator slItr1 = startLocs.begin();
         slItr1 != startLocs.end();
         ++slItr1 )
    {
      StartLoc* sl1 = *slItr1;

      map<StartLoc*, float>* innerMap = new map<StartLoc*, float>;

      float d1 = weightedInfluenceDistance( sl1, b );

      for( list<StartLoc*>::const_iterator slItr2 = startLocs.begin();
           slItr2 != startLocs.end();
           ++slItr2 )
      {
        StartLoc* sl2 = *slItr2;

        if( sl1 == sl2 ) { continue; }

        float d2 = weightedInfluenceDistance( sl2, b );

        float dTotal   = d1 + d2;
        float dAverage = 0.5f * dTotal;
        float pNeutral = 0.5f;

        float pInfl = pNeutral + ((dAverage - d1) / dTotal);

        innerMap->insert( make_pair( sl2, 100.0f * pInfl ) );

        sl1->sl2resourceInfluence[sl2] += b->resourceTotal * pInfl;
      }

      b->sl1vsl2_influence.insert( make_pair( sl1, innerMap ) );
    }
  }

  // calculate average influence for main/nat/third
  for( list<StartLoc*>::const_iterator slItr = startLocs.begin();
       slItr != startLocs.end();
       ++slItr )
  {
    StartLoc* sl = *slItr;

    for( list<Base*>::const_iterator bItr = bases.begin();
         bItr != bases.end();
         ++bItr )
    {
      Base* b = *bItr;

      b->sl2averageInfluence[sl] = averageInfluence( sl, b );
    }
  }
}


float SC2Map::weightedInfluenceDistance( StartLoc* sl, Base* b )
{
  // return a distance that is a weighted combination
  // of distances from this StartLoc to this Base
  float dwi = 0.0f;

  // when the distance to a base is infinite
  // like an island base by ground path, don't use
  // the infinity constant because it in the calculation
  // it completely obscures the non-infinite factors.
  // Instead, use an otherwise big number based on the
  // map's playable dimensions
  float dNoInfluence = (float) (txDimPlayable + tyDimPlayable);


  float dGround = getShortestPathDistance( &(sl->loc), &(b->loc), PATH_GROUND_WITHROCKS );
  if( SC2Map::effectivelyInfinity( dGround ) )
  {
    dGround = dNoInfluence;
  }
  dwi += influenceWeightGround * dGround;

  float dCWalk = getShortestPathDistance( &(sl->loc), &(b->loc), PATH_CWALK_WITHROCKS );
  if( SC2Map::effectivelyInfinity( dCWalk ) )
  {
    dCWalk = dNoInfluence;
  }
  dwi += influenceWeightCWalk * dCWalk;

  float dAir = getShortestAirDistance( &(sl->loc), &(b->loc) );
  dwi += influenceWeightAir * dAir;

  return dwi;
}


float SC2Map::averageInfluence( StartLoc* sl, Base* b )
{
  float total = 0.0f;

  for( list<StartLoc*>::const_iterator slItr = startLocs.begin();
       slItr != startLocs.end();
       ++slItr )
  {
    StartLoc* sl2 = *slItr;

    if( sl == sl2 ) { continue; }

    map<StartLoc*, float>* innerMap = b->sl1vsl2_influence[sl];
    total += (*innerMap)[sl2];
  }

  return total / (float)startLocs.size();
}


void SC2Map::assignMainNatThirdIslands()
{
  // find a start loc's main, nat and third
  // also calculate the list of expansions that are
  // an island relative to every start location

  // don't even consider a base as a main, nat or
  // third unless the start loc exerts an influence
  // greater than 100/#startLocs on the base
  float minInfluence = 100.0f / (float)startLocs.size();

  // start by adding all bases as potential islands
  islandExpos    .insert( bases.begin(), bases.end() );
  semiislandExpos.insert( bases.begin(), bases.end() );

  for( list<StartLoc*>::const_iterator slItr = startLocs.begin();
       slItr != startLocs.end();
       ++slItr )
  {
    StartLoc* sl = *slItr;

    sl->mainBase  = NULL;
    sl->natBase   = NULL;
    sl->thirdBase = NULL;

    for( list<Base*>::const_iterator bItr = bases.begin();
         bItr != bases.end();
         ++bItr )
    {
      Base* b = *bItr;

      // do island classifications
      float dGroundWR = getShortestPathDistance( &(sl->loc), &(b->loc), PATH_GROUND_WITHROCKS );
      float dGroundNR = getShortestPathDistance( &(sl->loc), &(b->loc), PATH_GROUND_NOROCKS   );

      // if a path without rocks, NOT AN ISLAND
      if( !SC2Map::effectivelyInfinity( dGroundNR ) )
      {
        islandExpos.erase( b );
      }

      // if a path with rocks, ALSO NOT A SEMIISLAND
      if( !SC2Map::effectivelyInfinity( dGroundWR ) )
      {
        semiislandExpos.erase( b );
      }

      // now do main/nat/third because we might jump out of this iteration
      float bInfluence = b->sl2averageInfluence[sl];

      if( bInfluence < minInfluence )
      {
        continue;
      }

      // should this base beat out the current third?
      if( sl->thirdBase == NULL ||
          sl->thirdBase->sl2averageInfluence[sl] < bInfluence )
      {
        sl->thirdBase = b;
      } else {
        continue;
      }

      // should this base beat out the current nat?
      if( sl->natBase == NULL ||
          sl->natBase->sl2averageInfluence[sl] < bInfluence )
      {
        sl->thirdBase = sl->natBase;
        sl->natBase   = b;
      } else {
        continue;
      }

      // should this base beat out the current main?
      if( sl->mainBase == NULL ||
          sl->mainBase->sl2averageInfluence[sl] < bInfluence )
      {
        sl->natBase  = sl->mainBase;
        sl->mainBase = b;
      }

    }
  }

  // if an expo is an island, it's not a semi island
  for( list<Base*>::const_iterator bItr = bases.begin();
       bItr != bases.end();
       ++bItr )
  {
    Base* b = *bItr;

    if( islandExpos.find( b ) != islandExpos.end() )
    {
      semiislandExpos.erase( b );
    }
  }
}


void SC2Map::calculatePositionalBalance()
{
  for( list<StartLoc*>::const_iterator slItr1 = startLocs.begin();
       slItr1 != startLocs.end();
       ++slItr1 )
  {
    StartLoc* sl1 = *slItr1;

    for( list<StartLoc*>::const_iterator slItr2 = startLocs.begin();
         slItr2 != startLocs.end();
         ++slItr2 )
    {
      StartLoc* sl2 = *slItr2;

      if( sl1 == sl2 ) { continue; }

      float r = sl1->sl2resourceInfluence[sl2] / (totalMinerals + totalVespeneGas);

      // multiply by 200 because we want locations to appear
      // balanced like this:  SL1 is 98% vs. SL2 at 102%
      sl1->sl2percentBalanced[sl2] = 200.0f * r;
    }
  }
}
