#include <stdlib.h>
#include <stdio.h>
#include <assert.h>

#include "outstreams.hpp"
#include "coordinates.hpp"
#include "SC2Map.hpp"


float SC2Map::OPENNESS_NOTCALCULATED = -1.0f;


void SC2Map::setOpennessRenderMax( float orm )
{
  opennessRenderMax = orm;
}

float SC2Map::opennessRenderMax = 15.0f;




  // openness is a numerical value for a cell that
  // determines how far away the nearest unpathable
  // cell is, or how "open" the cell is
void SC2Map::computeOpenness()
{
  // we need the previous pass's openness values
  // while we are updating the current values
  mapOpenness     = new float[cxDimPlayable*cyDimPlayable*NUM_PATH_TYPES];
  mapOpennessPrev = new float[cxDimPlayable*cyDimPlayable*NUM_PATH_TYPES];

  // initialize all cells for all pathing types to not-calculated
  for( int pci = 0; pci < cxDimPlayable; ++pci )
  {
    for( int pcj = 0; pcj < cyDimPlayable; ++pcj )
    {
      point c;
      c.pcSet( pci, pcj );

      setOpenness( &c, PATH_GROUND_NOROCKS,   OPENNESS_NOTCALCULATED );
      setOpenness( &c, PATH_GROUND_WITHROCKS, OPENNESS_NOTCALCULATED );
      setOpenness( &c, PATH_CWALK_NOROCKS,    OPENNESS_NOTCALCULATED );
      setOpenness( &c, PATH_CWALK_WITHROCKS,  OPENNESS_NOTCALCULATED );
    }
  }

  computeOpenness( PATH_GROUND_WITHROCKS );
  computeOpenness( PATH_GROUND_NOROCKS   );
}


void SC2Map::computeOpenness( PathType t )
{
  bool firstPass = true;

  // terminate algorithm when we've calculated every pathable cell
  int numCellsCalculated = 0;

  while( numCellsCalculated < numPathableCells[t] )
  {
    // to begin, copy current into the last pass
    memcpy( mapOpennessPrev,
            mapOpenness,
            cxDimPlayable*cyDimPlayable*NUM_PATH_TYPES*sizeof( float ) );

    // consult last pass to calculate current pass
    for( int pci = 0; pci < cxDimPlayable; ++pci )
    {
      for( int pcj = 0; pcj < cyDimPlayable; ++pcj )
      {
        point c;
        c.pcSet( pci, pcj );

        // only calculate openness for pathable cells
        if( !getPathing( &c, t ) )
        {
          continue;
        }

        // only set each cell once!
        if( checkHasOpennessLastPass( &c, t ) )
        {
          continue;
        }

        // Now we are considering a cell that has not been set, but IS pathable--
        // but, if no neighbors have been set, then it shouldn't be marked this pass
        //
        // If at least one neighbor has an openness value,
        // it should be set to the lowest of {openness value coming in
        // from a neighbor plus the distance to that neighbor}

        Node* u               = getPathNode( &c, t );
        float opennessCurrent = infinity;
        bool  markNow         = false;

        if( firstPass )
        {
          // at the base level (openness 0)
          // we test if any of the 4 cardinal
          // neighbors are unpathable
          if( u->neighbors[0] == NULL ||
              u->neighbors[1] == NULL ||
              u->neighbors[2] == NULL ||
              u->neighbors[3] == NULL )
          {
            markNow = true;
            opennessCurrent = 0.0f;
          }

        } else {
          // otherwise we test if any neighbors have had
          // their openness set, and if so take the lowest
          // openness value for the current cell
          for( int i = 0; i < NUM_NODE_NEIGHBORS; ++i )
          {
            Node* v = u->neighbors[i];
            if( v == NULL ) { continue; }

            if( checkHasOpennessLastPass( &(v->loc), t ) )
            {
              markNow = true;

              float openness = getOpennessLastPass( &(v->loc), t ) + neighborWeights[i];
              if( openness < opennessCurrent )
              {
                opennessCurrent = openness;
              }
            }
          }
        }

        if( markNow )
        {
          setOpenness( &c, t, opennessCurrent );

          if( opennessCurrent > opennessMax[t] )
          {
            opennessMax[t] = opennessCurrent;
          }

          ++numCellsCalculated;
        }
      }
    }

    if( firstPass )
    {
      firstPass = false;
    }
  }
}



void SC2Map::setOpenness( point* c, PathType t, float o )
{
  if( !isPlayableCell( c ) )
  {
    printError( "Attempt to access data from unplayable cell (%d, %d).\n",
                c->pcx,
                c->pcy );
    exit( -1 );
  }
  mapOpenness[NUM_PATH_TYPES*(c->pcy*cxDimPlayable + c->pcx) + t] = o;
}


bool SC2Map::checkHasOpenness( point* c, PathType t )
{
  if( !isPlayableCell( c ) )
  {
    printError( "Attempt to access data from unplayable cell (%d, %d)\n.",
                c->pcx,
                c->pcy );
    exit( -1 );
  }
  return mapOpenness[NUM_PATH_TYPES*(c->pcy*cxDimPlayable + c->pcx) + t] > -0.5f;
}


float SC2Map::getOpenness( point* c, PathType t )
{
  if( !isPlayableCell( c ) )
  {
    printError( "Attempt to access data from unplayable cell (%d, %d)\n.",
                c->pcx,
                c->pcy );
    exit( -1 );
  }
  return mapOpenness[NUM_PATH_TYPES*(c->pcy*cxDimPlayable + c->pcx) + t];
}



bool SC2Map::checkHasOpennessLastPass( point* c, PathType t )
{
  if( !isPlayableCell( c ) )
  {
    return false;
  }

  return
    mapOpennessPrev[NUM_PATH_TYPES*(c->pcy*cxDimPlayable + c->pcx) + t] > -0.5f;
}

float SC2Map::getOpennessLastPass( point* c, PathType t )
{
  if( !isPlayableCell( c ) )
  {
    printError( "Attempt to access data from unplayable cell (%d, %d)\n.",
                c->pcx,
                c->pcy );
    exit( -1 );
  }
  return mapOpennessPrev[NUM_PATH_TYPES*(c->pcy*cxDimPlayable + c->pcx) + t];
}
