#include <stdlib.h>
#include <stdio.h>
#include <assert.h>
#include <string.h>

#include "tinyxml.h"

#include "outstreams.hpp"
#include "coordinates.hpp"
#include "SC2Map.hpp"



float SC2Map::defaultMineralAmount = 1500.0f;
float SC2Map::defaultGeyserAmount  = 2500.0f;


// TODO
// <Point id="155" position="93.7172,88.8452,0" scale="1,1,1" type="BlockPathing" name="Block Pathing 001" color="0,0,0,0"/>
// <Unit id="90" position="86,72,0" rotation="5.496" scale="1,1,1" unitType="Observatory"/>


void SC2Map::processPlacedObject( TiXmlElement* object )
{
  if( strcmp( object->Value(), "Point" ) == 0 )
  {
    const char* strType = object->Attribute( "type" );
    if( strcmp( strType, "StartLoc" ) != 0 )
    {
      return;
    }

    const char* strPosition = object->Attribute( "position" );
    float mx, my, mz;
    sscanf( strPosition, "%f,%f,%f", &mx, &my, &mz );

    StartLoc* sl = new StartLoc;
    sl->loc.mSet( mx, my );
    startLocs.push_back( sl );
    return;
  }


  if( strcmp( object->Value(), "Unit" ) == 0 )
  {
    // every unit has a position (TRUE?)
    const char* strPosition = object->Attribute( "position" );
    float mx, my, mz;
    sscanf( strPosition, "%f,%f,%f", &mx, &my, &mz );

    // might be used by resources below
    const char* strResources = object->Attribute( "resources" );


    // every unit has a unitType (TRUE?)
    const char* strType = object->Attribute( "unitType" );

    if( strcmp( strType, "BraxisAlphaDestructible1x1" ) == 0 ||
        strcmp( strType, "PathingBlocker1x1Doodad"    ) == 0 )
    {
      point p;
      p.mSet( mx, my );
      processDestructible1x1( &p );
      return;
    }

    if( strcmp( strType, "BraxisAlphaDestructible2x2" ) == 0 )
    {
      point p;
      p.mSet( mx, my );
      processDestructible2x2( &p );
      return;
    }

    if( strcmp( strType, "DestructibleRock2x4Horizontal" ) == 0 )
    {
      point p;
      p.mSet( mx, my );
      processDestructible2x4Horizontal( &p );
      return;
    }

    if( strcmp( strType, "DestructibleRock2x4Vertical" ) == 0 )
    {
      point p;
      p.mSet( mx, my );
      processDestructible2x4Vertical( &p );
      return;
    }

    if( strcmp( strType, "DestructibleRock2x6Horizontal" ) == 0 )
    {
      point p;
      p.mSet( mx, my );
      processDestructible2x6Horizontal( &p );
      return;
    }

    if( strcmp( strType, "DestructibleRock2x6Vertical" ) == 0 )
    {
      point p;
      p.mSet( mx, my );
      processDestructible2x6Vertical( &p );
      return;
    }

    if( strcmp( strType, "DestructibleRock4x4"   ) == 0 ||
        strcmp( strType, "DestructibleDebris4x4" ) == 0 )
    {
      point p;
      p.mSet( mx, my );
      processDestructible4x4( &p );
      return;
    }

    if( strcmp( strType, "DestructibleRock6x6"   ) == 0 ||
        strcmp( strType, "DestructibleDebris6x6" ) == 0 )
    {
      point p;
      p.mSet( mx, my );
      processDestructible6x6( &p );
      return;
    }

    if( strcmp( strType, "DestructibleRampDiagonalHugeBLUR" ) == 0 )
    {
      point p;
      p.mSet( mx, my );
      processDestructibleHugeBLUR( &p );
      return;
    }

    if( strcmp( strType, "DestructibleRampDiagonalHugeULBR" ) == 0 )
    {
      point p;
      p.mSet( mx, my );
      processDestructibleHugeULBR( &p );
      return;
    }

    if( strcmp( strType, "MineralField" ) == 0 )
    {
      Resource* r = new Resource;
      r->loc.mSet( mx, my );
      r->cliffLevel = getHeight( &(r->loc) );
      r->type       = MINERALS;
      if( strResources != NULL )
      {
        r->amount = atof( strResources );
      } else {
        r->amount = defaultMineralAmount;
      }
      totalMinerals += r->amount;
      resources.push_back( r );
      return;
    }

    if( strcmp( strType, "HighYieldMineralField" ) == 0 )
    {
      Resource* r = new Resource;
      r->loc.mSet( mx, my );
      r->cliffLevel = getHeight( &(r->loc) );
      r->type       = MINERALS_HY;
      if( strResources != NULL )
      {
        r->amount = atof( strResources );
      } else {
        r->amount = defaultMineralAmount;
      }
      totalMinerals   += r->amount;
      totalHYMinerals += r->amount;
      resources.push_back( r );
      return;
    }

    if( strcmp( strType, "VespeneGeyser"       ) == 0 ||
        strcmp( strType, "SpacePlatformGeyser" ) == 0 )
    {
      Resource* r = new Resource;
      r->loc.mSet( mx, my );
      r->cliffLevel = getHeight( &(r->loc) );
      r->type       = VESPENEGAS;
      if( strResources != NULL )
      {
        r->amount = atof( strResources );
      } else {
        r->amount = defaultGeyserAmount;
      }
      totalVespeneGas += r->amount;
      resources.push_back( r );
      return;
    }

    if( strcmp( strType, "RichVespeneGeyser" ) == 0 )
    {
      Resource* r = new Resource;
      r->loc.mSet( mx, my );
      r->cliffLevel = getHeight( &(r->loc) );
      r->type       = VESPENEGAS_HY;
      if( strResources != NULL )
      {
        r->amount = atof( strResources );
      } else {
        r->amount = defaultGeyserAmount;
      }
      totalVespeneGas   += r->amount;
      totalHYVespeneGas += r->amount;
      resources.push_back( r );
      return;
    }

  }
}


void SC2Map::processDestructible1x1( point* p )
{
  removePathing( p );
}


void SC2Map::processDestructible2x2( point* p )
{
  point c0;  c0.pcSet ( p->pcx,   p->pcy   );
  point c1;  c1.pcSet ( p->pcx-1, p->pcy   );
  point c2;  c2.pcSet ( p->pcx,   p->pcy-1 );
  point c3;  c3.pcSet ( p->pcx-1, p->pcy-1 );

  removePathing( &c0  );
  removePathing( &c1  );
  removePathing( &c2  );
  removePathing( &c3  );
}

void SC2Map::processDestructible4x4( point* p )
{
  point c0;  c0.pcSet ( p->pcx,   p->pcy   );
  point c1;  c1.pcSet ( p->pcx-1, p->pcy   );
  point c2;  c2.pcSet ( p->pcx,   p->pcy-1 );
  point c3;  c3.pcSet ( p->pcx-1, p->pcy-1 );

  point c4;  c4.pcSet ( p->pcx-1, p->pcy+1 );
  point c5;  c5.pcSet ( p->pcx,   p->pcy+1 );
  point c6;  c6.pcSet ( p->pcx-1, p->pcy-2 );
  point c7;  c7.pcSet ( p->pcx,   p->pcy-2 );

  point c8;  c8.pcSet ( p->pcx+1, p->pcy-1 );
  point c9;  c9.pcSet ( p->pcx+1, p->pcy   );
  point c10; c10.pcSet( p->pcx-2, p->pcy-1 );
  point c11; c11.pcSet( p->pcx-2, p->pcy   );

  removePathing( &c0  );
  removePathing( &c1  );
  removePathing( &c2  );
  removePathing( &c3  );
  removePathing( &c4  );
  removePathing( &c5  );
  removePathing( &c6  );
  removePathing( &c7  );
  removePathing( &c8  );
  removePathing( &c9  );
  removePathing( &c10 );
  removePathing( &c11 );
}

void SC2Map::processDestructible6x6( point* p )
{
  point c0;  c0.pcSet ( p->pcx-2, p->pcy+1 );
  point c1;  c1.pcSet ( p->pcx-1, p->pcy+1 );
  point c2;  c2.pcSet ( p->pcx,   p->pcy+1 );
  point c3;  c3.pcSet ( p->pcx+1, p->pcy+1 );

  point c4;  c4.pcSet ( p->pcx-2, p->pcy   );
  point c5;  c5.pcSet ( p->pcx-1, p->pcy   );
  point c6;  c6.pcSet ( p->pcx,   p->pcy   );
  point c7;  c7.pcSet ( p->pcx+1, p->pcy   );

  point c8;  c8.pcSet ( p->pcx-2, p->pcy-1 );
  point c9;  c9.pcSet ( p->pcx-1, p->pcy-1 );
  point c10; c10.pcSet( p->pcx,   p->pcy-1 );
  point c11; c11.pcSet( p->pcx+1, p->pcy-1 );

  point c12; c12.pcSet( p->pcx-2, p->pcy-2 );
  point c13; c13.pcSet( p->pcx-1, p->pcy-2 );
  point c14; c14.pcSet( p->pcx,   p->pcy-2 );
  point c15; c15.pcSet( p->pcx+1, p->pcy-2 );

  point c16; c16.pcSet( p->pcx-1, p->pcy+2 );
  point c17; c17.pcSet( p->pcx,   p->pcy+2 );

  point c18; c18.pcSet( p->pcx-1, p->pcy-3 );
  point c19; c19.pcSet( p->pcx,   p->pcy-3 );

  point c20; c20.pcSet( p->pcx+2, p->pcy-1 );
  point c21; c21.pcSet( p->pcx+2, p->pcy   );

  point c22; c22.pcSet( p->pcx-3, p->pcy-1 );
  point c23; c23.pcSet( p->pcx-3, p->pcy   );

  removePathing( &c0  );
  removePathing( &c1  );
  removePathing( &c2  );
  removePathing( &c3  );
  removePathing( &c4  );
  removePathing( &c5  );
  removePathing( &c6  );
  removePathing( &c7  );
  removePathing( &c8  );
  removePathing( &c9  );
  removePathing( &c10 );
  removePathing( &c11 );
  removePathing( &c12 );
  removePathing( &c13 );
  removePathing( &c14 );
  removePathing( &c15 );
  removePathing( &c16 );
  removePathing( &c17 );
  removePathing( &c18 );
  removePathing( &c19 );
  removePathing( &c20 );
  removePathing( &c21 );
  removePathing( &c22 );
  removePathing( &c23 );
}

void SC2Map::processDestructible2x4Horizontal( point* p )
{
  point c0;  c0.pcSet ( p->pcx,   p->pcy   );
  point c1;  c1.pcSet ( p->pcx-1, p->pcy   );
  point c2;  c2.pcSet ( p->pcx,   p->pcy-1 );
  point c3;  c3.pcSet ( p->pcx-1, p->pcy-1 );

  point c8;  c8.pcSet ( p->pcx+1, p->pcy-1 );
  point c9;  c9.pcSet ( p->pcx+1, p->pcy   );
  point c10; c10.pcSet( p->pcx-2, p->pcy-1 );
  point c11; c11.pcSet( p->pcx-2, p->pcy   );

  removePathing( &c0  );
  removePathing( &c1  );
  removePathing( &c2  );
  removePathing( &c3  );
  removePathing( &c8  );
  removePathing( &c9  );
  removePathing( &c10 );
  removePathing( &c11 );
}

void SC2Map::processDestructible2x4Vertical( point* p )
{
  point c0;  c0.pcSet ( p->pcx,   p->pcy   );
  point c1;  c1.pcSet ( p->pcx-1, p->pcy   );
  point c2;  c2.pcSet ( p->pcx,   p->pcy-1 );
  point c3;  c3.pcSet ( p->pcx-1, p->pcy-1 );

  point c4;  c4.pcSet ( p->pcx-1, p->pcy+1 );
  point c5;  c5.pcSet ( p->pcx,   p->pcy+1 );
  point c6;  c6.pcSet ( p->pcx-1, p->pcy-2 );
  point c7;  c7.pcSet ( p->pcx,   p->pcy-2 );

  removePathing( &c0  );
  removePathing( &c1  );
  removePathing( &c2  );
  removePathing( &c3  );
  removePathing( &c4  );
  removePathing( &c5  );
  removePathing( &c6  );
  removePathing( &c7  );
}



void SC2Map::processDestructible2x6Horizontal( point* p )
{
  point c0;  c0.pcSet ( p->pcx,   p->pcy   );
  point c1;  c1.pcSet ( p->pcx-1, p->pcy   );
  point c2;  c2.pcSet ( p->pcx,   p->pcy-1 );
  point c3;  c3.pcSet ( p->pcx-1, p->pcy-1 );

  point c8;  c8.pcSet ( p->pcx+1, p->pcy-1 );
  point c9;  c9.pcSet ( p->pcx+1, p->pcy   );
  point c10; c10.pcSet( p->pcx-2, p->pcy-1 );
  point c11; c11.pcSet( p->pcx-2, p->pcy   );

  point c20; c20.pcSet( p->pcx+2, p->pcy-1 );
  point c21; c21.pcSet( p->pcx+2, p->pcy   );
  point c22; c22.pcSet( p->pcx-3, p->pcy-1 );
  point c23; c23.pcSet( p->pcx-3, p->pcy   );

  removePathing( &c0  );
  removePathing( &c1  );
  removePathing( &c2  );
  removePathing( &c3  );
  removePathing( &c8  );
  removePathing( &c9  );
  removePathing( &c10 );
  removePathing( &c11 );
  removePathing( &c20 );
  removePathing( &c21 );
  removePathing( &c22 );
  removePathing( &c23 );
}

void SC2Map::processDestructible2x6Vertical( point* p )
{
  point c0;  c0.pcSet ( p->pcx,   p->pcy   );
  point c1;  c1.pcSet ( p->pcx-1, p->pcy   );
  point c2;  c2.pcSet ( p->pcx,   p->pcy-1 );
  point c3;  c3.pcSet ( p->pcx-1, p->pcy-1 );

  point c4;  c4.pcSet ( p->pcx-1, p->pcy+1 );
  point c5;  c5.pcSet ( p->pcx,   p->pcy+1 );
  point c6;  c6.pcSet ( p->pcx-1, p->pcy-2 );
  point c7;  c7.pcSet ( p->pcx,   p->pcy-2 );

  point c20; c20.pcSet( p->pcx-1, p->pcy+2 );
  point c21; c21.pcSet( p->pcx,   p->pcy+2 );
  point c22; c22.pcSet( p->pcx-1, p->pcy-3 );
  point c23; c23.pcSet( p->pcx,   p->pcy-3 );

  removePathing( &c0  );
  removePathing( &c1  );
  removePathing( &c2  );
  removePathing( &c3  );
  removePathing( &c4  );
  removePathing( &c5  );
  removePathing( &c6  );
  removePathing( &c7  );
  removePathing( &c20 );
  removePathing( &c21 );
  removePathing( &c22 );
  removePathing( &c23 );
}




////////////////////////////////////////////////
//
//  The huge destructible UL to BR:
//
//     ##
//   ##@@##
//     ##@@##
//       ##@@##
//         ##@@CC
//           ##@@##
//             ##@@##
//               ##@@##
//                 ##@@##
//                   ##
//
////////////////////////////////////////////////
void SC2Map::processDestructibleHugeULBR( point* p )
{
  point c0;  c0.pcSet ( p->pcx-4, p->pcy+4 );
  point c1;  c1.pcSet ( p->pcx-5, p->pcy+3 );
  point c2;  c2.pcSet ( p->pcx-4, p->pcy+3 );
  point c3;  c3.pcSet ( p->pcx-3, p->pcy+3 );
  point c4;  c4.pcSet ( p->pcx-4, p->pcy+2 );
  point c5;  c5.pcSet ( p->pcx-3, p->pcy+2 );
  point c6;  c6.pcSet ( p->pcx-2, p->pcy+2 );
  point c7;  c7.pcSet ( p->pcx-3, p->pcy+1 );
  point c8;  c8.pcSet ( p->pcx-2, p->pcy+1 );
  point c9;  c9.pcSet ( p->pcx-1, p->pcy+1 );
  point c10; c10.pcSet( p->pcx-2, p->pcy   );
  point c11; c11.pcSet( p->pcx-1, p->pcy   );
  point c12; c12.pcSet( p->pcx,   p->pcy   );
  point c13; c13.pcSet( p->pcx-1, p->pcy-1 );
  point c14; c14.pcSet( p->pcx,   p->pcy-1 );
  point c15; c15.pcSet( p->pcx+1, p->pcy-1 );
  point c16; c16.pcSet( p->pcx,   p->pcy-2 );
  point c17; c17.pcSet( p->pcx+1, p->pcy-2 );
  point c18; c18.pcSet( p->pcx+2, p->pcy-2 );
  point c19; c19.pcSet( p->pcx+1, p->pcy-3 );
  point c20; c20.pcSet( p->pcx+2, p->pcy-3 );
  point c21; c21.pcSet( p->pcx+3, p->pcy-3 );
  point c22; c22.pcSet( p->pcx+2, p->pcy-4 );
  point c23; c23.pcSet( p->pcx+3, p->pcy-4 );
  point c24; c24.pcSet( p->pcx+4, p->pcy-4 );
  point c25; c25.pcSet( p->pcx+3, p->pcy-5 );

  removePathing( &c0  );
  removePathing( &c1  );
  removePathing( &c2  );
  removePathing( &c3  );
  removePathing( &c4  );
  removePathing( &c5  );
  removePathing( &c6  );
  removePathing( &c7  );
  removePathing( &c8  );
  removePathing( &c9  );
  removePathing( &c10 );
  removePathing( &c11 );
  removePathing( &c12 );
  removePathing( &c13 );
  removePathing( &c14 );
  removePathing( &c15 );
  removePathing( &c16 );
  removePathing( &c17 );
  removePathing( &c18 );
  removePathing( &c19 );
  removePathing( &c20 );
  removePathing( &c21 );
  removePathing( &c22 );
  removePathing( &c23 );
  removePathing( &c24 );
  removePathing( &c25 );
}

////////////////////////////////////////////////
//
//  The huge destructible BL to UR:
//
//                   ##
//                 ##@@##
//               ##@@##
//             ##@@##
//           ##CC##
//         ##@@##
//       ##@@##
//     ##@@##
//   ##@@##
//     ##
//
////////////////////////////////////////////////
void SC2Map::processDestructibleHugeBLUR( point* p )
{
  point c0;  c0.pcSet ( p->pcx+3, p->pcy+4 );
  point c1;  c1.pcSet ( p->pcx+2, p->pcy+3 );
  point c2;  c2.pcSet ( p->pcx+3, p->pcy+3 );
  point c3;  c3.pcSet ( p->pcx+4, p->pcy+3 );
  point c4;  c4.pcSet ( p->pcx+1, p->pcy+2 );
  point c5;  c5.pcSet ( p->pcx+2, p->pcy+2 );
  point c6;  c6.pcSet ( p->pcx+3, p->pcy+2 );
  point c7;  c7.pcSet ( p->pcx,   p->pcy+1 );
  point c8;  c8.pcSet ( p->pcx+1, p->pcy+1 );
  point c9;  c9.pcSet ( p->pcx+2, p->pcy+1 );
  point c10; c10.pcSet( p->pcx-1, p->pcy   );
  point c11; c11.pcSet( p->pcx,   p->pcy   );
  point c12; c12.pcSet( p->pcx+1, p->pcy   );
  point c13; c13.pcSet( p->pcx-2, p->pcy-1 );
  point c14; c14.pcSet( p->pcx-1, p->pcy-1 );
  point c15; c15.pcSet( p->pcx,   p->pcy-1 );
  point c16; c16.pcSet( p->pcx-3, p->pcy-2 );
  point c17; c17.pcSet( p->pcx-2, p->pcy-2 );
  point c18; c18.pcSet( p->pcx-1, p->pcy-2 );
  point c19; c19.pcSet( p->pcx-4, p->pcy-3 );
  point c20; c20.pcSet( p->pcx-3, p->pcy-3 );
  point c21; c21.pcSet( p->pcx-2, p->pcy-3 );
  point c22; c22.pcSet( p->pcx-5, p->pcy-4 );
  point c23; c23.pcSet( p->pcx-4, p->pcy-4 );
  point c24; c24.pcSet( p->pcx-3, p->pcy-4 );
  point c25; c25.pcSet( p->pcx-4, p->pcy-5 );

  removePathing( &c0  );
  removePathing( &c1  );
  removePathing( &c2  );
  removePathing( &c3  );
  removePathing( &c4  );
  removePathing( &c5  );
  removePathing( &c6  );
  removePathing( &c7  );
  removePathing( &c8  );
  removePathing( &c9  );
  removePathing( &c10 );
  removePathing( &c11 );
  removePathing( &c12 );
  removePathing( &c13 );
  removePathing( &c14 );
  removePathing( &c15 );
  removePathing( &c16 );
  removePathing( &c17 );
  removePathing( &c18 );
  removePathing( &c19 );
  removePathing( &c20 );
  removePathing( &c21 );
  removePathing( &c22 );
  removePathing( &c23 );
  removePathing( &c24 );
  removePathing( &c25 );
}



void SC2Map::removePathing( point* c )
{
  removePathing( c, PATH_GROUND_WITHROCKS );
  removePathing( c, PATH_CWALK_WITHROCKS  );
}


void SC2Map::removePathing( point* c, PathType t )
{
  if( getPathing( c, t ) )
  {
    numPathableCells[t]--;
    setPathing( c, t, false );
  }
}
