#include <stdlib.h>
#include <math.h>

#include "pngwriter.h"

#include "outstreams.hpp"
#include "coordinates.hpp"
#include "SC2Map.hpp"



// constants relating to the image frame
int SC2Map::iDimT = 8;
int SC2Map::iDimC = 6;

/////////////////////////////////////////////////
//
//  NOTE!  The image frame is really a canvas
//  of pixels, but in the final out PNG there
//  will be borders and text outside the canvas.
//  Don't include these borders in the base
//  coordinate translations because we want to be
//  able to adjust the borders as we decide how
//  to render the image--and probably render
//  several images, each with different possible
//  border settings.
//
/////////////////////////////////////////////////

int SC2Map::iLeft   = 0;
int SC2Map::iBottom = 0;
int SC2Map::iRight  = 0;
int SC2Map::iTop    = 0;

// use this to take a point.ix -> outPNG coordinates

// note that the grid of terrain is shifted by half a
// unit to the cell grid, and we let the playable
// terrain surround the playable cells, so there is an
// extra half-unit of terrain going all the way around
// the cell grid, account for this in the shift
int SC2Map::ix2png( int ix )
{
  return iLeft + 1 + iDimT/2 + ix;
}

int SC2Map::iy2png( int iy )
{
  return iBottom + 1 + iDimT/2 + iy;
}


void SC2Map::renderFooter()
{
  plotText( "generated by SC2 Map Analyzer, http://www.sc2mapster.com/assets/sc2-map-analyzer",
            14,          // font size
            false,       // centered,
            20, 20,  // bottom-left
            0.0f,        // angle
            1.0f, 1.0f, 1.0f, // foreground
            0.0f, 0.0f, 0.0f );
}


void SC2Map::renderImageTerrainOnly()
{
  resetPNGWriter( "-terrain-only",
                  20, 20,
                  58,
                  44 );

  char title[2*FILENAME_LENGTH];
  sprintf( title,
           "%s - Terrain only",
           this->strMapName );

  plotText( title,
            18,          // font size
            false,       // centered,
            20, iTop+20,  // bottom-left
            0.0f,        // angle
            1.0f, 1.0f, 1.0f, // foreground
            0.0f, 0.0f, 0.0f );

  renderTerrain();
  renderResources();
  renderStartLocs();
  renderFooter();
  img->close();
}


void SC2Map::renderImagesPathing()
{
  renderImagePathing( "-pathingGroundWithRocks",    PATH_GROUND_WITHROCKS );
  renderImagePathing( "-pathingGroundNoRocks",      PATH_GROUND_NOROCKS   );
  renderImagePathing( "-pathingCliffWalkWithRocks", PATH_CWALK_WITHROCKS  );
  renderImagePathing( "-pathingCliffWalkNoRocks",   PATH_CWALK_NOROCKS    );
}

void SC2Map::renderImagePathing( const char* strOutPNG,
                                 PathType    t )
{
  resetPNGWriter( strOutPNG,
                  20, 20,
                  58,
                  44 );

  char strPath[64];
  switch( t )
  {
    case PATH_GROUND_WITHROCKS:
    {
      sprintf( strPath, "Ground pathing, with rocks" );
    } break;

    case PATH_GROUND_NOROCKS:
    {
      sprintf( strPath, "Ground pathing, without rocks" );
    } break;

    case PATH_CWALK_WITHROCKS:
    {
      sprintf( strPath, "Cliff-walk pathing, with rocks" );
    } break;

    case PATH_CWALK_NOROCKS:
    {
      sprintf( strPath, "Cliff-walk pathing, without rocks" );
    } break;

    default:
      sprintf( strPath, "UNKNOWN PATHING TYPE?" );
  }

  char title[2*FILENAME_LENGTH];
  sprintf( title,
           "%s - %s",
           this->strMapName,
           strPath );

  plotText( title,
            18,          // font size
            false,       // centered,
            20, iTop+20,  // bottom-left
            0.0f,        // angle
            1.0f, 1.0f, 1.0f, // foreground
            0.0f, 0.0f, 0.0f );

  renderTerrain();
  renderPathing( t );
  renderResources();
  renderStartLocs();
  renderFooter();
  img->close();
}


void SC2Map::renderImageBases()
{
  resetPNGWriter( "-bases",
                  20, 20,
                  58,
                  44 );

  char title[2*FILENAME_LENGTH];
  sprintf( title,
           "%s - Bases identified",
           this->strMapName );

  plotText( title,
            18,          // font size
            false,       // centered,
            20, iTop+20,  // bottom-left
            0.0f,        // angle
            1.0f, 1.0f, 1.0f, // foreground
            0.0f, 0.0f, 0.0f );

  renderTerrain();
  renderOpenness( PATH_GROUND_WITHROCKS );
  renderResources();
  renderBases();
  renderFooter();
  img->close();
}


void SC2Map::renderImagesShortestPaths()
{
  for( list<StartLoc*>::const_iterator itr1 = startLocs.begin();
       itr1 != startLocs.end();
       ++itr1 )
  {
    StartLoc* sl1 = *itr1;

    list<StartLoc*>::const_iterator itr2 = itr1;
    ++itr2;
    for( ;
         itr2 != startLocs.end();
         ++itr2 )
    {
      StartLoc* sl2 = *itr2;

      char filename[FILENAME_LENGTH];
      sprintf( filename, "-shortestPaths-%d-%d", sl1->idNum, sl2->idNum );

      char title[2*FILENAME_LENGTH];
      sprintf( title,
               "%s - Shortest paths, %s to %s",
               this->strMapName,
               sl1->name,
               sl2->name );

      renderImageShortestPaths( filename,
                                title,
                                &(sl1->loc),
                                &(sl2->loc) );

      if( sl1->natBase != NULL && sl2->natBase != NULL )
      {
        sprintf( filename, "-shortestPaths-%d-%d-nats", sl1->idNum, sl2->idNum );

        sprintf( title,
                 "%s - Shortest paths, naturals, %s to %s",
                 this->strMapName,
                 sl1->name,
                 sl2->name );

        renderImageShortestPaths( filename,
                                  title,
                                  &(sl1->natBase->loc),
                                  &(sl2->natBase->loc) );
      }
    }
  }
}

void SC2Map::renderImageShortestPaths( char*  filename,
                                       char*  title,
                                       point* p0,
                                       point* p1 )
{
  resetPNGWriter( filename,
                  20, 20, // borders left right
                  86,     // border top
                  44 );   // border bottom


  plotText( title,
            18,          // font size
            false,       // centered,
            20, iTop+48, // bottom-left
            0.0f,        // angle
            1.0f, 1.0f, 1.0f, // foreground
            0.0f, 0.0f, 0.0f );

  int textOffset = 20;

  char path[32];

  sprintf( path,
           "by Air  " );
  plotText( path,
            18,          // font size
            false,       // centered,
            20 + textOffset, iTop+20, // bottom-left
            0.0f,        // angle
            0.0f, 1.0f, 1.0f, // foreground
            0.0f, 0.0f, 0.0f );

  textOffset += getTextWidth( path, 18 );

  sprintf( path,
           "by Ground  " );
  plotText( path,
            18,          // font size
            false,       // centered,
            20 + textOffset, iTop+20, // bottom-left
            0.0f,        // angle
            1.0f, 1.0f, 0.0f, // foreground
            0.0f, 0.0f, 0.0f );

  textOffset += getTextWidth( path, 18 );

  sprintf( path,
           "by Cliff-Walk" );
  plotText( path,
            18,          // font size
            false,       // centered,
            20 + textOffset, iTop+20, // bottom-left
            0.0f,        // angle
            1.0f, 0.0f, 1.0f, // foreground
            0.0f, 0.0f, 0.0f );


  renderTerrain();
  renderPathing( PATH_GROUND_WITHROCKS );
  renderResources();

  plotAirPath( p0, p1,
               0.0f, 1.0f, 1.0f,
               0.25f );

  plotShortestPath( p0, p1, PATH_GROUND_WITHROCKS,
                    1.0f, 1.0f, 0.0f,
                    0.50f );

  plotShortestPath( p0, p1, PATH_CWALK_WITHROCKS,
                    1.0f, 0.0f, 1.0f,
                    0.75f );

  renderStartLocs();
  renderFooter();
  img->close();
}

void SC2Map::renderImageOpenness()
{
  resetPNGWriter( "-openness",
                  20, 20, // borders left right
                  86,     // border top
                  44 );   // border bottom

  char title[2*FILENAME_LENGTH];
  sprintf( title,
           "%s - Openness, by ground pathing with rocks",
           this->strMapName );

  plotText( title,
            18,          // font size
            false,       // centered,
            20, iTop+48, // bottom-left
            0.0f,        // angle
            1.0f, 1.0f, 1.0f, // foreground
            0.0f, 0.0f, 0.0f );

  float r, g, b;

  for( int i = 0; i < 255; ++i )
  {
    float fi       = (float)i;
    float openness = opennessRenderMax * fi / 200.0f;

    openness2color( openness, &r, &g, &b );

    img->line( 30+i, iTop+18,
               30+i, iTop+40,
               r, g, b );
  }

  char num[32];

  openness2color( 0.0f, &r, &g, &b );
  sprintf( num, "%.1f", 0.0f );
  plotText( num,
            16,          // font size
            true,        // centered,
            50, iTop+30, //
            0.0f,        // angle
            r,    g,    b, // foreground
            0.0f, 0.0f, 0.0f );

  openness2color( opennessRenderMax*0.5f, &r, &g, &b );
  sprintf( num, "%.1f", opennessRenderMax*0.5f );
  plotText( num,
            16,          // font size
            true,        // centered,
            150, iTop+30, //
            0.0f,        // angle
            r,    g,    b, // foreground
            0.0f, 0.0f, 0.0f );

  openness2color( opennessRenderMax, &r, &g, &b );
  sprintf( num, "%.1f+", opennessRenderMax );
  plotText( num,
            16,          // font size
            true,        // centered,
            250, iTop+30, //
            0.0f,        // angle
            r,    g,    b, // foreground
            0.0f, 0.0f, 0.0f );

  renderTerrain();
  renderOpenness( PATH_GROUND_WITHROCKS );
  renderResources();
  renderStartLocs();
  renderFooter();
  img->close();
}


void SC2Map::renderImagesInfluence()
{
  for( list<StartLoc*>::const_iterator itr1 = startLocs.begin();
       itr1 != startLocs.end();
       ++itr1 )
  {
    StartLoc* sl1 = *itr1;

    list<StartLoc*>::const_iterator itr2 = itr1;
    ++itr2;
    for( ;
         itr2 != startLocs.end();
         ++itr2 )
    {
      StartLoc* sl2 = *itr2;

      char filename[FILENAME_LENGTH];
      sprintf( filename, "-influence-%d-%d", sl1->idNum, sl2->idNum );

      char title[2*FILENAME_LENGTH];
      sprintf( title,
               "%s - Influence, %s vs %s",
               this->strMapName,
               sl1->name,
               sl2->name );

      renderImageInfluence( filename,
                            title,
                            sl1,
                            sl2 );
    }
  }
}

void SC2Map::renderImageInfluence( char*     filename,
                                   char*     title,
                                   StartLoc* sl1,
                                   StartLoc* sl2 )
{
  resetPNGWriter( filename,
                  20, 20, // borders: left, right
                  86,     // border: top
                  44      // border: bottom
                  );

  plotText( title,
            18,          // font size
            false,       // centered,
            20, iTop+48, // bottom-left
            0.0f,        // angle
            1.0f, 1.0f, 1.0f, // foreground
            0.0f, 0.0f, 0.0f );

  char balance[2*FILENAME_LENGTH];
  sprintf( balance,
           "Positional Balance: %s(%.1f%%) vs %s(%.1f%%)",
           sl1->name,
           sl1->sl2percentBalanced[sl2],
           sl2->name,
           sl2->sl2percentBalanced[sl1] );

  plotText( balance,
            18,          // font size
            false,       // centered,
            20, iTop+20, // bottom-left
            0.0f,        // angle
            1.0f, 1.0f, 1.0f, // foreground
            0.0f, 0.0f, 0.0f );

  renderTerrain();
  renderOpenness( PATH_GROUND_WITHROCKS );
  renderResources();
  renderInfluence( sl1, sl2 );
  plotStartLoc( sl1 );
  plotStartLoc( sl2 );
  renderFooter();
  img->close();
}

void SC2Map::renderTestImages()
{
  //resetPNGWriter( "-test-frame-translation",
  //                3, 7, // left, right borders
  //                11, 4  // top, bottom border
  //              );
  //renderTestFrameTranslation();
  //img->close();

  resetPNGWriter( "-test-clock-face-detection",
                  5, 5, // left, right borders
                  5, 5  // top, bottom border
                );
  renderTestClockFaceDetection();
  img->close();
}


void SC2Map::resetPNGWriter( const char* strFileSuffix,
                             int         iBorderLeft,
                             int         iBorderRight,
                             int         iBorderTop,
                             int         iBorderBottom )
{
  char strFilename[FILENAME_LENGTH];

  sprintf( strFilename,
           "%s%s.png",
           this->strMapFilename,
           strFileSuffix );

  img->pngwriter_rename( strFilename );

  // this resize of the image buffer implicitly colors
  // all pixels black
  img->resize( txDimPlayable*iDimT + iBorderLeft   + iBorderRight,
               tyDimPlayable*iDimT + iBorderBottom + iBorderTop );

  iLeft   = iBorderLeft;
  iBottom = iBorderBottom;
  iRight  = iBorderLeft   + txDimPlayable*iDimT;
  iTop    = iBorderBottom + tyDimPlayable*iDimT;
}



void SC2Map::renderTerrain()
{
  for( int pti = 0; pti < txDimPlayable; ++pti )
  {
    for( int ptj = 0; ptj < tyDimPlayable; ++ptj )
    {
      point t;
      t.ptSet( pti, ptj );
      plotTerrain( &t );
    }
  }
}

void SC2Map::plotTerrain( point* t )
{
  float r, g, b;

  float h0 = 0.00f;
  float h1 = 0.15f;
  float h2 = 0.25f;
  float h3 = 0.35f;

  switch( getHeight( t ) )
  {
    case 0x0: r = h0; g = h0; b = h0; break;
    case 0x1: r = h1; g = h1; b = h1; break;
    case 0x2: r = h2; g = h2; b = h2; break;
    case 0x3: r = h3; g = h3; b = h3; break;
  }

  plotTerrainPoint( t, r, g, b );
}




void SC2Map::renderPathing( PathType t )
{
  for( int pci = 0; pci < cxDimPlayable; ++pci )
  {
    for( int pcj = 0; pcj < cyDimPlayable; ++pcj )
    {
      point c;
      c.pcSet( pci, pcj );
      plotPathing( &c, t );
    }
  }
}

void SC2Map::plotPathing( point* c, PathType t )
{
  float r, g, b;

  if( getPathing( c, t ) )
  {
    r = 0.0f; g = 0.7f; b = 0.0f;
  } else {
    r = 0.7f; g = 0.0f; b = 0.0f;
  }

  plotCellPoint( c, r, g, b );
}



void SC2Map::renderOpenness( PathType t )
{
  for( int pci = 0; pci < cxDimPlayable; ++pci )
  {
    for( int pcj = 0; pcj < cyDimPlayable; ++pcj )
    {
      point c;
      c.pcSet( pci, pcj );
      plotOpenness( &c, t );
    }
  }
}


void SC2Map::plotOpenness( point* c, PathType t )
{
  float r, g, b;

  if( !checkHasOpenness( c, t ) )
  {
    r = 0.0f; g = 0.0f; b = 0.0f;

  } else {

    float openness = getOpenness( c, t );
    openness2color( openness, &r, &g, &b );
  }

  plotCellPoint( c, r, g, b );
}


void SC2Map::openness2color( float opennessIn,
                             float* rOut, float* gOut, float* bOut )
{
  float d = 2.0f * opennessIn / opennessRenderMax;

  if( d > 1.0f )
  {
    d -= 1.0f;
    *rOut = 0.3f;
    *gOut = 1.0f - d*0.8f;
    *bOut = 0.1f + d*0.8f;
  } else {
    *rOut = 0.3f + (1.0f-d)*0.7f;
    *gOut = 0.3f + d*0.7f;
    *bOut = 0.1f;
  }
}


void SC2Map::renderResources()
{
  for( list<Resource*>::const_iterator itr = resources.begin();
       itr != resources.end();
       ++itr )
  {
    Resource* r = *itr;
    switch( r->type )
    {
      case MINERALS:
        plotMineralField( &(r->loc), false );
        break;

      case MINERALS_HY:
        plotMineralField( &(r->loc), true );
        break;

      case VESPENEGAS:
      case VESPENEGAS_HY:
        plotGeyser( &(r->loc) );
        break;

      default:
        printError( "Unknown resource type.\n" );
        exit( -1 );
    }
  }
}

void SC2Map::plotMineralField( point* p, bool highYield )
{
  float r, g, b;

  if( highYield )
  {
    r = 0.8f; g = 0.8f; b = 0.0f;
  } else {
    r = 0.0f; g = 7.0f; b = 1.0f;
  }

  int x0 = ix2png( p->ix - iDimT   );
  int y0 = iy2png( p->iy - iDimT/2 );

  img->filledsquare( x0,
                     y0,
                     x0 + 2*iDimT - 1,
                     y0 +   iDimT - 1,
                     r, g, b
                   );
}

void SC2Map::plotGeyser( point* p )
{
  int x0 = ix2png( p->ix - 3*(iDimT/2) );
  int y0 = iy2png( p->iy - 3*(iDimT/2) );

  img->filledsquare( x0,
                     y0,
                     x0 + 3*iDimT - 1,
                     y0 + 3*iDimT - 1,
                     1.0f, 1.0f, 1.0f
                   );
}



void SC2Map::renderStartLocs()
{
  for( list<StartLoc*>::const_iterator itr = startLocs.begin();
       itr != startLocs.end();
       ++itr )
  {
    plotStartLoc( *itr );
  }
}

void SC2Map::plotStartLoc( StartLoc* sl )
{
  float r = 1.0f;
  float g = 1.0f;
  float b = 1.0f;

  int x0 = ix2png( sl->loc.ix - 5*(iDimT/2) );
  int y0 = iy2png( sl->loc.iy - 5*(iDimT/2) );

  img->filledsquare( x0,
                     y0,
                     x0 + 5*iDimT - 1,
                     y0 + 5*iDimT - 1,
                     r, g, b
                   );

  plotText( sl->name, 18,
            true,               // centered
            ix2png( sl->loc.ix ),
            iy2png( sl->loc.iy ),
            0.0f,               // angle
            1.0f, 1.0f, 1.0f,   // foreground
            0.0f, 0.0f, 0.0f ); // background
}




void SC2Map::renderBases()
{
  for( list<Base*>::const_iterator itr = bases.begin();
       itr != bases.end();
       ++itr )
  {
    plotBase( *itr );
  }
}

void SC2Map::plotBase( Base* base )
{
  // for aligning text to plot
  int classOffset = -24;
  int totalOffset = -48;

  // draw arrow pointing to resource patches that
  // belong to this base
  for( list<Resource*>::const_iterator itr = base->resources.begin();
       itr != base->resources.end();
       ++itr )
  {
    Resource* r = *itr;

    plotArrow( ix2png( base->loc.ix ), iy2png( base->loc.iy ),
               ix2png( r->loc.ix    ), iy2png( r->loc.iy    ),
               1.0f, 0.0f, 0.0f );
  }

  int x0 = ix2png( base->loc.ix - (iDimT/2) );
  int y0 = iy2png( base->loc.iy - (iDimT/2) );

  img->filledsquare( x0,
                     y0,
                     x0 + iDimT - 1,
                     y0 + iDimT - 1,
                     1.0f, 0.0f, 0.0f
                   );

  // write the classification of the base
  if( semiislandExpos.find( base ) != semiislandExpos.end() )
  {
    plotText( "Semi Island",
              18,         // font size
              true,       // centered,
              ix2png( base->loc.ix ), iy2png( base->loc.iy + classOffset ),
              0.0f,        // angle
              0.2f, 0.6f, 1.0f, // foreground
              0.0f, 0.0f, 0.0f );
  }

  if( islandExpos.find( base ) != islandExpos.end() )
  {
    plotText( "Island",
              18,         // font size
              true,       // centered,
              ix2png( base->loc.ix ), iy2png( base->loc.iy + classOffset ),
              0.0f,        // angle
              0.2f, 0.6f, 1.0f, // foreground
              0.0f, 0.0f, 0.0f );
  }

  for( list<StartLoc*>::const_iterator itr = startLocs.begin();
       itr != startLocs.end();
       ++itr )
  {
    StartLoc* sl = *itr;

    if( base == sl->mainBase )
    {
      plotText( "Main",
                18,         // font size
                true,       // centered,
                ix2png( base->loc.ix ), iy2png( base->loc.iy + classOffset ),
                0.0f,        // angle
                1.0f, 1.0f, 0.8f, // foreground
                0.0f, 0.0f, 0.0f );

    } else if( base == sl->natBase ) {
      plotText( "Natural",
                18,         // font size
                true,       // centered,
                ix2png( base->loc.ix ), iy2png( base->loc.iy + classOffset ),
                0.0f,        // angle
                0.7f, 1.0f, 0.7f, // foreground
                0.0f, 0.0f, 0.0f );

    } else if( base == sl->thirdBase ) {
      plotText( "Third",
                18,         // font size
                true,       // centered,
                ix2png( base->loc.ix ), iy2png( base->loc.iy + classOffset ),
                0.0f,        // angle
                1.0f, 0.7f, 0.7f, // foreground
                0.0f, 0.0f, 0.0f );
    }
  }

  char strResources[64];
  sprintf( strResources, "%.1fK", base->resourceTotal / 1000.0f );
  plotText( strResources,
            14,         // font size
            true,       // centered,
            ix2png( base->loc.ix ), iy2png( base->loc.iy + totalOffset ),
            0.0f,        // angle
            1.0f, 1.0f, 1.0f, // foreground
            0.0f, 0.0f, 0.0f );
}



void SC2Map::renderInfluence( StartLoc* sl1, StartLoc* sl2 )
{
  for( list<Base*>::const_iterator itr = bases.begin();
       itr != bases.end();
       ++itr )
  {
    Base* b = *itr;

    map<StartLoc*, float>* innerMap = b->sl1vsl2_influence[sl1];
    float infl1 = (*innerMap)[sl2];

    innerMap = b->sl1vsl2_influence[sl2];
    float infl2 = (*innerMap)[sl1];

    char strInfl[64];
    sprintf( strInfl,
             "%.1f%%v%.1f%%",
             infl1,
             infl2);

    plotText( strInfl,
              18,          // font size
              true,       // centered,
              ix2png( b->loc.ix ), iy2png( b->loc.iy ),
              0.0f,        // angle
              1.0f, 1.0f, 1.0f, // foreground
              0.0f, 0.0f, 0.0f );
  }
}



void SC2Map::plotAirPath( point* p0, point* p1,
                          float r, float g, float b,
                          float labelPlacement )
{
  plotLine( ix2png( p0->ix ),
            iy2png( p0->iy ),
            ix2png( p1->ix ),
            iy2png( p1->iy ),
            2.0f,
            r, g, b );

  float dTotal = getShortestAirDistance( p0, p1 );

  float dix = (float)(p1->ix - p0->ix);
  float diy = (float)(p1->iy - p0->iy);

  int ixLabel = p0->ix + (int)(dix*labelPlacement);
  int iyLabel = p0->iy + (int)(diy*labelPlacement);

  char strDistance[512];
  sprintf( strDistance, "%.1f", dTotal );

  plotText( strDistance, 18,
            true,               // centered
            ix2png( ixLabel ),
            iy2png( iyLabel ),
            0.0f,               // angle
            r,    g,    b,      // foreground
            0.0f, 0.0f, 0.0f ); // background
}




void SC2Map::plotShortestPath( point* p0, point* p1, PathType t,
                               float r, float g, float b,
                               float labelPlacement )
{
  Node* src = getPathNode( p0, t );
  Node* u   = getPathNode( p1, t );
  Node* v   = getShortestPathPredecessor( src, u, t );

  bool positionedLabel = false;

  float dTotal = getShortestPathDistance( src, u, t );

  int ixLabel;
  int iyLabel;

  while( v != NULL )
  {
    plotLine( ix2png( u->loc.ix ), iy2png( u->loc.iy ),
              ix2png( v->loc.ix ), iy2png( v->loc.iy ),
              2.0f,
              r, g, b );
    u = v;
    v = getShortestPathPredecessor( src, u, t );

    float d = getShortestPathDistance( src, u, t );
    if( !positionedLabel && d < dTotal * labelPlacement )
    {
      ixLabel = u->loc.ix;
      iyLabel = u->loc.iy;
      positionedLabel = true;
    }
  }

  char strDistance[512];
  sprintf( strDistance, "%.1f", dTotal );

  plotText( strDistance, 18,
            true,               // centered
            ix2png( ixLabel ),
            iy2png( iyLabel ),
            0.0f,               // angle
            r,    g,    b,      // foreground
            0.0f, 0.0f, 0.0f ); // background
}




void SC2Map::plotAllShortestPaths( int sl, PathType t )
{
  /*
  float r = 1.0f;
  float g = 0.0f;
  float b = 1.0f;

  for( int pci = 0; pci < cxDimPlayable; ++pci )
  {
    for( int pcj = 0; pcj < cyDimPlayable; ++pcj )
    {
      point c; c.pcSet( pci, pcj );
      Node* u = getPathNode( &c, t );
      if( u == NULL ) { continue; }

      Node* v = u->pi[sl];
      if( v == NULL ) { continue; }

      img->arrow( u->loc.ix, u->loc.iy,
                  v->loc.ix, v->loc.iy,
                  4, 0.4f,
                  r, g, b );
    }
  }
  */
}






void SC2Map::plotTerrainPoint( point* t, float r, float g, float b )
{
  // Terrain point looks like, where ## is the
  // bottom-left pixel of the bounding box
  //
  //        t1t1
  //    t1t1t1t1t1t1
  //    t1t1t1t1t1t1
  //  t1t1t1t1t1t1t1t1
  //  t1t1t1t1t1t1t1t1
  //    t1t1t1t1t1t1
  //    t1t1t1t1t1t1
  //  ##    t1t1


  // get the bottom-left corner of the terrain point
  // in output PNG pixels
  int x0 = ix2png( t->ix - iDimT/2 );
  int y0 = iy2png( t->iy - iDimT/2 );

  // do most of the thing as a rectangle--remember
  // pngwriter plots the inclsive rectangle, so subtract
  // one from the top and right coordinates
  img->filledsquare( x0 + 1,
                     y0 + 1,
                     x0 + iDimT - 2,
                     y0 + iDimT - 2,
                     r, g, b
                   );

  // then plot the 8 pixels that stick out of the
  // four sides of the bigger rectable
  img->plot( x0,             y0 + 3,         r, g, b );
  img->plot( x0,             y0 + 4,         r, g, b );
  img->plot( x0 + iDimT - 1, y0 + 3,         r, g, b );
  img->plot( x0 + iDimT - 1, y0 + 4,         r, g, b );
  img->plot( x0 + 3,         y0,             r, g, b );
  img->plot( x0 + 4,         y0,             r, g, b );
  img->plot( x0 + 3,         y0 + iDimT - 1, r, g, b );
  img->plot( x0 + 4,         y0 + iDimT - 1, r, g, b );
}



void SC2Map::plotCellPoint( point* c, float r, float g, float b )
{
  // get the bottom-left corner of the cell
  // in output PNG pixels
  int x0 = ix2png( c->ix - iDimC/2 );
  int y0 = iy2png( c->iy - iDimC/2 );

  img->filledsquare( x0,
                     y0 + 2,
                     x0 + 5,
                     y0 + 3,
                     r, g, b
                   );

  img->filledsquare( x0 + 2,
                     y0,
                     x0 + 3,
                     y0 + 5,
                     r, g, b
                   );
}



void SC2Map::plotArrow( int x0, int y0, int x1, int y1,
                        float r, float g, float b )
{
  if( x0 == x1 && y0 == y1 )
  {
    // apparently if you try to plot an arrow from one
    // point to the same point, pngwriter dies silently
    return;
  }

  img->arrow( x0, y0, x1, y1,
              4, 0.4f, // size and angle of arrowhead
              r, g, b );
}



void SC2Map::plotLine( int x0, int y0, int x1, int y1,
                       float thick,
                       float r, float g, float b )
{
  ////////////////////////////////////////
  //
  //        x10,y10
  //             .
  //             x1,y1
  //              /| .
  //             .   x11, y11
  //   x00,y00  . ang
  //        .  .
  //        x0,y0
  //            .
  //            x01,y01
  //
  ////////////////////////////////////////

  float ang = atan2( y1-y0, x1-x0 ) + M_PI*0.5f;

  int dx = (int)(thick*cos( ang ));
  int dy = (int)(thick*sin( ang ));

  int x00 = x0 + dx;
  int y00 = y0 + dy;
  int x01 = x0 - dx;
  int y01 = y0 - dy;

  int x10 = x1 + dx;
  int y10 = y1 + dy;
  int x11 = x1 - dx;
  int y11 = y1 - dy;

  plotTri( x00, y00,
           x01, y01,
           x11, y11,
           r, g, b );

  plotTri( x00, y00,
           x10, y10,
           x11, y11,
           r, g, b );
}



void SC2Map::plotTri( int x1, int y1,
                      int x2, int y2,
                      int x3, int y3,
                      float r, float g, float b )
{
  // for some reason it appears sorting the
  // three points by ascending y coordinate
  // prevents nasty visual artifacts

  int xa, ya;
  int xb, yb;
  int xc, yc;

  if( y1 < y2 ) {
    // y1 < y2

    if( y2 < y3 ) {
      // y1 < y2 < y3
      xa = x1; ya = y1;
      xb = x2; yb = y2;
      xc = x3; yc = y3;

    } else {
      // y1 < y2   y3 < y2

      if( y1 < y3 ) {
        // y1 < y3 < y2
        xa = x1; ya = y1;
        xb = x3; yb = y3;
        xc = x2; yc = y2;

      } else {
        // y3 < y1 < y2
        xa = x3; ya = y3;
        xb = x1; yb = y1;
        xc = x2; yc = y2;
      }
    }

  } else {
    // y2 < y1

    if( y1 < y3 ) {
      // y2 < y1 < y3
      xa = x2; ya = y2;
      xb = x1; yb = y1;
      xc = x3; yc = y3;

    } else {
      // y2 < y1   y3 < y1

      if( y2 < y3 ) {
        // y2 < y3 < y1
        xa = x2; ya = y2;
        xb = x3; yb = y3;
        xc = x1; yc = y1;

      } else {
        // y3 < y2 < y1
        xa = x3; ya = y3;
        xb = x2; yb = y2;
        xc = x1; yc = y1;
      }
    }

  }

  img->filledtriangle( xa, ya,
                       xb, yb,
                       xc, yc,
                       r, g, b );
}


void SC2Map::plotText( char* text,
                       int fontsize,
                       bool centered, int ix, int iy,
                       float angle,
                       float rf, float gf, float bf, // foreground
                       float rb, float gb, float bb  // background
                     )
{
  char fontface[] = "./FreeSansBold.ttf";

  int x;
  int y;

  if( centered )
  {
    int w = img->get_text_width( fontface, fontsize, text );

    x = ix - w/2;
    y = iy - fontsize/2;

  } else {
    // othewise its just the lower left corner
    x = ix;
    y = iy;
  }

  // make a background of the text with offset plots
  for( int i = -2; i <= 2; ++i )
  {
    for( int j = -2; j <= 2; ++j )
    {
      img->plot_text( fontface, fontsize,
                      x + i, y + j,
                      angle, text,
                      rb, gb, bb );
    }
  }

  // then put foreground down
  img->plot_text( fontface, fontsize,
                  x, y,
                  angle, text,
                  rf, gf, bf );
}


int SC2Map::getTextWidth( char* text, int fontsize )
{
  char fontface[] = "./FreeSansBold.ttf";
  return img->get_text_width( fontface, fontsize, text );
}




void SC2Map::renderTestFrameTranslation()
{
  float r, g, b;

  img->filledsquare( iLeft, iBottom, iRight - 1, iTop - 1, 0.2f, 0.2f, 0.0f );


  renderTerrain();

  r = 1.0f; g = 0.0f; b = 0.0f;
  img->plot( ix2png( 0 ), iy2png( 0 ), r, g, b );
  //plotText( "image frame origin", 10, true, iLeft, iBottom-10,
  //          0.0f, r, g, b, 0.0f, 0.0f, 0.0f );

  // when the map (mcLeft, mcBottom) = (26, 20) then
  // this point should appear in the center of the
  // playable cell (0, 0)
  r = 0.0f; g = 1.0f; b = 0.0f;
  point p1; p1.mSet( 26.5f, 20.5f );
  img->plot( ix2png( p1.ix ), iy2png( p1.iy ), r, g, b );


  // this point should appear in the center of
  // playable terrain unit (1, 2)
  r = 0.0f; g = 0.0f; b = 1.0f;
  point p2; p2.ptSet( 1, 2 );
  img->plot( ix2png( p2.ix ), iy2png( p2.iy ), r, g, b );

  // and playable terrain unit (2, 2)
  p2.ptSet( 2, 2 );
  img->plot( ix2png( p2.ix ), iy2png( p2.iy ), r, g, b );

  // and playable terrain unit (1, 3)
  p2.ptSet( 1, 3 );
  img->plot( ix2png( p2.ix ), iy2png( p2.iy ), r, g, b );



  // this point should appear in the center of
  // playable cell (3, 1)
  r = 0.0f; g = 1.0f; b = 1.0f;
  point p3; p3.pcSet( 3, 1 );
  img->plot( ix2png( p3.ix ), iy2png( p3.iy ), r, g, b );

  // and playable cell (4, 1)
  p3.pcSet( 4, 1 );
  img->plot( ix2png( p3.ix ), iy2png( p3.iy ), r, g, b );

  // and playable cell (5, 2)
  p3.pcSet( 5, 2 );
  img->plot( ix2png( p3.ix ), iy2png( p3.iy ), r, g, b );

  p3.pcSet( 0, 0 );
  img->plot( ix2png( p3.ix ), iy2png( p3.iy ), r, g, b );
}


void SC2Map::renderTestClockFaceDetection()
{
  float upperBounds[13];

  // each wedge of clock is 1/12 of 2*pi, or pi/6
  float wedge = M_PI / 6.0f;

  // the upper bound of 9 o'clock's BOTTOM half is half a wedge
  // over the bottom of the range
  float bound = -M_PI + 0.5f*wedge;

  for( int i = 0; i < 13; ++i )
  {
    upperBounds[i] = bound;
    bound += wedge;
  }

  // start at bottom of range, go positive radians around clock
  float wedgeColors[][3] =
  {
    { 1.0f, 0.0f, 0.0f }, // 9
    { 0.0f, 1.0f, 0.0f }, // 8
    { 0.0f, 0.0f, 1.0f }, // 7
    { 1.0f, 1.0f, 0.0f }, // 6
    { 0.0f, 1.0f, 1.0f }, // 5
    { 1.0f, 0.0f, 1.0f }, // 4
    { 1.0f, 0.0f, 0.0f }, // 3
    { 0.0f, 1.0f, 0.0f }, // 2
    { 0.0f, 0.0f, 1.0f }, // 1
    { 1.0f, 1.0f, 0.0f }, // 12
    { 0.0f, 1.0f, 1.0f }, // 11
    { 1.0f, 0.0f, 1.0f }, // 10
    { 1.0f, 0.0f, 0.0f }, // 9
  };

  // set a point at the playable terrain's center
  // to calculate the angle to a start location
  point center;
  center.ptSet( txDimPlayable/2,
                tyDimPlayable/2 );

  for( int pci = 0; pci < cxDimPlayable; ++pci )
  {
    for( int pcj = 0; pcj < cyDimPlayable; ++pcj )
    {
      point c;
      c.pcSet( pci, pcj );

      float* rgb;

      // TODO alter angle for aspect ratio?
      float ang = atan2( c.my - center.my, c.mx - center.mx );

      for( int i = 0; i < 13; ++i )
      {
        if( ang < upperBounds[i] )
        {
          rgb = wedgeColors[i];
          break;
        }
      }

      plotCellPoint( &c, rgb[0], rgb[1], rgb[2] );
    }
  }
}
