#include <stdlib.h>
#include <stdio.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <unistd.h>
#include <string.h>
#include <dirent.h>

#include "outstreams.hpp"
#include "SC2Map.hpp"
#include "SC2MapAggregator.hpp"


// controls set by options: defaults
static bool pauseTerminal   = true;
static bool dirRecurse      = false;
static bool renderTerrain   = false;
static bool renderPathing   = false;
static bool renderBases     = false;
static bool renderOpenness  = false;
static bool renderShortest  = false;
static bool renderInfluence = false;
static bool writeCSVmaps    = false;
static bool writeCSVaggr    = false;
static bool defaultOutput   = true;


void optsFail();
void statFail( const char* filename );

void tryFileOrDir( const char* filename, bool recurse );
void tryFile     ( const char* filename );


void usage()
{
  printMessage( "\n" );
  printMessage( "usage: sc2mapanalyzer FILES\n" );
  printMessage( "  If FILE is a directory, all maps inside are analyzed.\n" );
  printMessage( "\n" );
  printMessage( "Use -r to recursively explore all directory arguments.\n" );
  printMessage( "Use -p to prevent the command window from staying open after analysis.\n" );
  printMessage( "Use -h to print a brief description of available options.\n" );
  printMessage( "\n" );
  printMessage( "For the complete manual, visit:\n" );
  printMessage( "http://www.sc2mapster.com/assets/sc2-map-analyzer\n" );
}


void help()
{
  printMessage( "\n" );
  printMessage( "usage: sc2mapanalyzer FILES\n" );
  printMessage( "  If FILE is a directory, all maps inside are analyzed.\n" );
  printMessage( "\n" );
  printMessage( "-h  Display this help message.\n" );
  printMessage( "\n" );
  printMessage( "-p  Prevent the terminal from pausing open after analysis.\n" );
  printMessage( "\n" );
  printMessage( "-r  Recursively search for maps in any directory arguments.\n" );
  printMessage( "\n" );
  printMessage( "-m  Set the openness render max, default is 15.0 (in map cells).\n" );
  printMessage( "\n" );
  printMessage( "-g -c -a  Set ground/cliff-walk/air influence weights, should add up to 1.\n" );
  printMessage( "\n" );
  printMessage( "-t  Render PNG of terrain only.\n" );
  printMessage( "\n" );
  printMessage( "-n  Render PNGs of pathing data.\n" );
  printMessage( "\n" );
  printMessage( "-b  Render PNG of bases.\n" );
  printMessage( "\n" );
  printMessage( "-o  Render PNG of openness.\n" );
  printMessage( "\n" );
  printMessage( "-s  Render PNGs of shortest paths.\n" );
  printMessage( "\n" );
  printMessage( "-i  Render PNGs of influence calculations.\n" );
  printMessage( "\n" );
  printMessage( "-v  Write a CSV summarizing all maps.\n" );
  printMessage( "\n" );
  printMessage( "-w  Write a CSV for each map's details.\n" );
  printMessage( "\n" );
  printMessage( "-e  Enable all outputs described above.\n" );
  printMessage( "\n" );
}


void AtExit()
{
  // flush output
  fflush( stdout );
  fflush( stderr );

  // if warnings or errors were reported,
  // make sure to hold terminal open so
  // user can read them
  if( pauseTerminal && warningOrErrorReported )
  {
    printMessage( "\nTo prevent this pause at termination, pass option -p\n." );
    system( "pause" );
  }
}


static SC2MapAggregator mapAggregator;


int main( int argc, char** argv )
{
  atexit( &AtExit );

  // from unistd.h for getopt
  opterr = 0;

  int c;
  while( (c = getopt( argc, argv, "hprm:g:c:a:tnbosivwe" )) != -1 )
  {
    switch( c )
    {
      case 'h':
        help();
        exit( 0 );
        break;

      case 'p':
        pauseTerminal = false;
        break;

      case 'r':
        dirRecurse = true;
        break;

      case 'm':
        SC2Map::setOpennessRenderMax( atof( optarg ) );
        break;

      case 'g':
        SC2Map::setInfluenceWeightGround( atof( optarg ) );
        break;

      case 'c':
        SC2Map::setInfluenceWeightCWalk( atof( optarg ) );
        break;

      case 'a':
        SC2Map::setInfluenceWeightAir( atof( optarg ) );
        break;

      case 't':
        renderTerrain = true;
        defaultOutput = false;
        break;

      case 'n':
        renderPathing = true;
        defaultOutput = false;
        break;

      case 'b':
        renderBases   = true;
        defaultOutput = false;
        break;

      case 'o':
        renderOpenness = true;
        defaultOutput  = false;
        break;

      case 's':
        renderShortest = true;
        defaultOutput  = false;
        break;

      case 'i':
        renderInfluence = true;
        defaultOutput   = false;
        break;

      case 'v':
        writeCSVaggr  = true;
        defaultOutput = false;
        break;

      case 'w':
        writeCSVmaps  = true;
        defaultOutput = false;
        break;

      case 'e':
        renderTerrain   = true;
        renderPathing   = true;
        renderBases     = true;
        renderOpenness  = true;
        renderShortest  = true;
        renderInfluence = true;
        writeCSVaggr    = true;
        writeCSVmaps    = true;
        defaultOutput   = false;
        break;

      case ':':
      case '?':
        // bad options, inform user
        optsFail();
        // no break, fall through

      default:
        usage();
        exit( -1 );
    }
  }


  // if no output was specified, enable
  // the default output
  if( defaultOutput )
  {
    renderTerrain   = false;
    renderPathing   = false;
    renderBases     = true;
    renderOpenness  = true;
    renderShortest  = true;
    renderInfluence = true;
    writeCSVaggr    = true;
    writeCSVmaps    = true;
  }


  // if, after options, there are no arguments
  // then search the current directory for maps
  if( optind == argc )
  {
    printMessage( "No file arguments, searching current directory for maps...\n" );
    tryFileOrDir( ".", true );
  }

  // process non-option arguments
  for( int i = optind; i < argc; ++i )
  {
    tryFileOrDir( argv[i], true );
  }

  if( writeCSVaggr )
  {
    mapAggregator.buildOutputColumns();
    mapAggregator.writeToCSV( "aggregate-summary.csv" );
  }

  return 0;
}


void optsFail()
{
  if( optopt == 'm' ||
      optopt == 'g' ||
      optopt == 'c' ||
      optopt == 'a' )
  {
    printError( "Option -%c requires an argument.\n", optopt );
  }

  else if (isprint (optopt))
  {
    printError( "Unknown option `-%c'.\n", optopt );

  } else {
    fprintf (stderr,
           "Unknown option character `\\x%x'.\n",
           optopt);
         }
}



void statFail( const char* filename )
{
  switch( errno )
  {
    case EACCES:
      printWarning( "Permission denied for %s, skipping...\n",
                    filename );
      break;

    case EBADF:
    case EFAULT:
      printWarning( "%s is an invalid file, skipping...\n",
                    filename );
      break;

    case ENAMETOOLONG:
      printWarning( "Path and filename %s is too long, skipping...\n",
                    filename );
      break;

    case ENOENT:
    case ENOTDIR:
      printWarning( "A component of the path %s does not exist or is not a directory, skipping...\n",
                    filename );
      break;

    case ENOMEM:
      printError( "Not enough system memory to continue.\n" );
      exit( -1 );
      break;

    default:
      printWarning( "Unknown problem trying to access %s, skipping...\n",
                    filename );
  }
}


// find out if the non-option argument is a file or
// a directory, and for directories decide if we should
// recursively explore them if the flag is set
void tryFileOrDir( const char* filenameIn, bool recurse )
{
  // make a copy of the filename so we can muck around with it
  char filename[FILENAME_LENGTH];
  memset( filename, '\0', FILENAME_LENGTH );
  strncpy( filename, filenameIn, FILENAME_LENGTH - 1 );

  // if the filename ends in a "/" or "\" we will interpret
  // it as a directory, but the trailing slash wigs out the
  // call to stat so trim it here
  if( filename[strlen( filename ) - 1] == '/' ||
      filename[strlen( filename ) - 1] == '\\' )
  {
    filename[strlen( filename ) - 1] = '\0';
  }

  // find out what kind of file we have
  struct stat fileStatus;
  if( stat( filename, &fileStatus ) < 0 )
  {
    statFail( filename );
    return;
  }

  // only deal with regular files or directories,
  // ignore other file types (or "modes")
  if( S_ISREG( fileStatus.st_mode ) )
  {
    tryFile( filename );


  } else if( S_ISDIR( fileStatus.st_mode ) ) {

    if( !recurse )
    {
      return;
    }

    DIR* dir = opendir( filename );
    if( dir != NULL )
    {
      struct dirent* de;
      while( de = readdir( dir ) )
      {
        // ignore the current and up-one directories, . and ..
        if( strcmp( de->d_name, "."  ) == 0 ||
            strcmp( de->d_name, ".." ) == 0 )
        {
          continue;
        }

        // make a path out of this directory name
        char filePath[FILENAME_LENGTH];
        strcpy( filePath, filename );
        int len = strlen( filePath );
        filePath[len]     = '\\';
        filePath[len + 1] = '\0';

        char* filenameFull = strcat( filePath, de->d_name );

        // only continue to recurse beyond the first level
        // if the user passed recursive option
        tryFileOrDir( filenameFull, dirRecurse );
      }
      closedir( dir );

    } else {
      printWarning( "Could not open directory %s.\n", filename );
    }

  } else {
    printWarning( "%s is not a regular file or directory, skipping...", filename );
  }
}


// found a regular file, see if it has a map
// extension and then try to open the archive
void tryFile( const char* filename )
{
  char* ext1 = strstr( filename, ".s2ma"  );
  char* ext2 = strstr( filename, ".SC2Map" );

  if( ext1 == NULL && ext2 == NULL )
  {
    // not a Starcraft 2 Map Archive...
    return;
  }

  printMessage( "Opening %s...\n", filename );

  SC2Map* sc2map = new SC2Map();

  if( sc2map->readMap( filename ) < 0 )
  {
    printWarning( "Could not read required map files for %s, skipping\n", filename );
    return;
  }

  printMessage( "Prepping shortest paths...\n" );
  sc2map->prepShortestPaths();

  printMessage( "Identifying bases...\n" );
  sc2map->identifyBases();

  printMessage( "Computing openness...\n" );
  sc2map->computeOpenness();

  printMessage( "Analyzing Bases...\n" );
  sc2map->analyzeBases();

  printMessage( "Generating output...\n" );

  if( renderTerrain )
  {
    sc2map->renderImageTerrainOnly();
  }

  if( renderPathing )
  {
    sc2map->renderImagesPathing();
  }

  if( renderBases )
  {
    sc2map->renderImageBases();
  }

  if( renderOpenness )
  {
    sc2map->renderImageOpenness();
  }

  if( renderShortest )
  {
    sc2map->renderImagesShortestPaths();
  }

  if( renderInfluence )
  {
    sc2map->renderImagesInfluence();
  }

  if( writeCSVmaps )
  {
    sc2map->writeToCSV( "-summary.csv" );
  }

  if( writeCSVaggr )
  {
    mapAggregator.aggregate( sc2map );
  }

  //sc2map->renderTestImages();

  delete sc2map;
}
